from __future__ import annotations

import math
from typing import Any, Dict, List

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import torch


def save_plot_history(history: Dict[str, List[float]], plot: bool = False, save_path: str | None = None) -> None:
    sns.set_theme(style="whitegrid")
    plt.rcParams["font.family"] = "DejaVu Sans"

    num_metrics = len(history.keys())
    num_cols = 2
    num_rows = math.ceil(num_metrics / num_cols)

    fig, axs = plt.subplots(num_rows, num_cols, figsize=(15, 10))

    axs = axs.flatten()

    for i, metric in enumerate(history.keys()):
        axs[i].plot(history[metric])
        axs[i].set_title(metric)
        axs[i].xaxis.set_major_locator(plt.MaxNLocator(integer=True))

    # Remove unused subplots
    for i in range(num_metrics, len(axs)):
        fig.delaxes(axs[i])

    plt.tight_layout()

    if plot:
        plt.show()  # type: ignore[no-untyped-call]

    if save_path is not None:
        plt.savefig(save_path, bbox_inches="tight", dpi=300)


def show_attention_heatmaps(
    attention_weights: torch.Tensor,
    xlabel: str = "Keys",
    ylabel: str = "Queries",
    xticks: List[str] | None = None,
    yticks: List[str] | None = None,
    show_title: bool = False,
    show_values: bool = False,
    value_dp: int = 2,
    figure_kwargs: Dict[str, Any] | None = None,
    plot_kwargs: Dict[str, Any] | None = None,
) -> plt.Figure:
    """
    Visualizes attention heatmaps for a given attention weight tensor.

    Parameters
    ----------
    attention_weights : torch.Tensor
        Attention weight tensor of shape (B, H, _, _).
    xlabel : str, optional
        Label for the x-axis. Default is "Keys".
    ylabel : str, optional
        Label for the y-axis. Default is "Queries".
    xticks : list of str, optional
        Labels for the x-axis ticks. If provided, must have the same length as the number of keys.
    yticks : list of str, optional
        Labels for the y-axis ticks. If provided, must have the same length as the number of queries.
    show_title : bool, optional
        Whether to show titles for each subplot. Default is False.
    show_values : bool, optional
        Whether to display attention weight values on the heatmap. Default is False.
    value_dp : int, optional
        Number of decimal places to display for the attention weight values. Default is 2.
    figure_kwargs : dict, optional
        Additional keyword arguments for the figure creation.
    plot_kwargs : dict, optional
        Additional keyword arguments for the heatmap plot.

    Returns
    -------
    fig : plt.Figure
        The figure object containing the attention heatmaps.

    Raises
    ------
    TypeError
        If `attention_weights` is not a PyTorch tensor.
    ValueError
        If `attention_weights` does not have shape (B, H, _, _).
    """
    if not isinstance(attention_weights, torch.Tensor):
        raise TypeError(f"Attention weights must be a PyTorch tensor, but got {type(attention_weights)}.")

    if len(attention_weights.shape) != 4:
        raise ValueError(f"Attention weights must have shape (B, H, _, _), but got {attention_weights.shape}.")

    B, H, num_queries, num_keys = attention_weights.shape

    if xticks is not None and len(xticks) != num_keys:
        raise ValueError(f"Length of xticks must match the number of keys. Expected {num_keys}, but got {len(xticks)}.")

    if yticks is not None and len(yticks) != num_queries:
        raise ValueError(
            f"Length of yticks must match the number of queries. Expected {num_queries}, but got {len(yticks)}."
        )

    attention_weights = attention_weights.cpu().detach().numpy()

    figure_kwargs = figure_kwargs or {"figsize": (15, 15), "sharex": True, "sharey": True, "squeeze": False}
    fig, axes = plt.subplots(B, H, **figure_kwargs)

    plot_kwargs = plot_kwargs or {"cmap": "viridis"}
    if B == 1:
        attention_weight = attention_weights[0]
        for h, (ax, head_attention) in enumerate(zip(axes, attention_weight)):
            sns.heatmap(head_attention, ax=ax, annot=show_values, fmt=f".{value_dp}f", **plot_kwargs)
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)

            if xticks is not None:
                ax.set_xticks(np.arange(num_keys) + 0.5)
                ax.set_xticklabels(xticks, rotation=45, ha="right")
            if yticks is not None:
                ax.set_yticks(np.arange(num_queries) + 0.5)
                ax.set_yticklabels(yticks)
            if show_title:
                ax.set_title(f"Head {h + 1}")
    else:
        for b, (row_axes, attention_weight) in enumerate(zip(axes, attention_weights)):
            for h, (ax, head_attention) in enumerate(zip(row_axes, attention_weight)):
                sns.heatmap(head_attention, ax=ax, annot=show_values, fmt=f".{value_dp}f", **plot_kwargs)
                if b == B - 1:
                    ax.set_xlabel(xlabel)  # Only the last batch will have the xlabel
                if h == 0:
                    ax.set_ylabel(ylabel)  # Only the first head will have the ylabel

                if xticks is not None:
                    ax.set_xticks(np.arange(num_keys) + 0.5)
                    ax.set_xticklabels(xticks, rotation=45, ha="right")
                if yticks is not None:
                    ax.set_yticks(np.arange(num_queries) + 0.5)
                    ax.set_yticklabels(yticks)

                if show_title:
                    ax.set_title(f"Batch {b + 1}, Head {h + 1}")

    fig.subplots_adjust(wspace=0.2, hspace=0.2)  # Adjust subplot spacing

    plt.show()  # type: ignore[no-untyped-call]

    return fig



from abc import ABC, abstractmethod
from typing import Optional, Tuple

import torch
from torch import nn


# FIXME: consider removing ABC since nn.Module ensures forward method to be implemented.
class Attention(ABC, nn.Module):
    """
    Base class for attention mechanisms.

    This abstract class provides a scaffold for attention mechanisms, with a
    dropout layer for regularization included. Subclasses are expected to
    implement the `forward` method.

    Attributes
    ----------
    dropout : The dropout layer applied to the attention scores.
        type: nn.Dropout

    Note
    ----
    ABC method might be redundant since inheritance from nn.Module ensures
    forward method to be implemented.
    """

    def __init__(self, dropout: float = 0.0) -> None:
        super().__init__()
        self.dropout = nn.Dropout(p=dropout, inplace=False)

    @abstractmethod
    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        mask: Optional[torch.BoolTensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Perform the forward pass for the attention mechanism."""
        raise NotImplementedError("The `forward` method must be implemented by the subclass.")



from __future__ import annotations

from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Type, Union

import torch
from pydantic import BaseModel, Field, field_validator, model_validator

from omnivault.transformer.utils.general_utils import PYTORCH_DTYPE_MAP
from omnivault.utils.torch_utils.device import get_device

__all__ = ["TrainerConfig"]


class TrainerConfig(BaseModel):
    device: torch.device = Field(default_factory=get_device, description="Device to use for training.")

    # general
    max_epochs: int = Field(default=2, description="Number of epochs to train for.")
    log_every_n_steps: int = Field(default=1, description="Log every n steps.")
    eval_every_n_steps: int = Field(default=1, description="Number of epochs between evaluations.")
    step_scheduler_on_batch_or_epoch: str = Field(
        default="epoch",
        description="Whether to step the scheduler on batch or epoch. "
        "If set to 'epoch', the scheduler will be stepped after each epoch. "
        "If set to 'batch', the scheduler will be stepped after each batch.",
    )

    # mixed precision training
    use_amp: bool = Field(default=False, description="Whether to use automatic mixed precision training.")
    autocast_config: Dict[str, Any] = Field(
        default={"enabled": False, "dtype": None, "cache_enabled": None},
        description="Autocast configuration, for details of the params, see `torch.cuda.amp.autocast`.",
    )
    scaler_config: Dict[str, Any] = Field(
        default={
            "enabled": False,
            "init_scale": 2.0**16,
            "growth_factor": 2.0,
            "backoff_factor": 0.5,
            "growth_interval": 2000,
        },
        description="Grad scaler configuration, for details of the params, see `torch.cuda.amp.GradScaler`. If not enabled, it is a no ops.",
    )

    # gradient accumulation
    gradient_accumulation_steps: int = Field(
        default=1, description="Number of gradient accumulation steps before performing a backward/update pass."
    )

    # training stability
    # 1. gradient clipping
    clip_grad_norm: Union[Dict[str, Any], None] = Field(
        default={"max_norm": 1.0, "norm_type": 2.0, "error_if_nonfinite": False, "foreach": None},
        description="Gradient clipping, for details of the params, see `torch.nn.utils.clip_grad_norm_`.",
    )

    # 2. weight decay on targetted parameter groups
    apply_weight_decay_to_different_param_groups: bool = Field(
        default=False, description="Whether to apply weight decay to different parameter groups."
    )

    # saving shenanigans
    save_dir: str = Field(default="checkpoints", description="Directory to save checkpoints to.")
    save_every_epoch: bool = Field(default=False, description="Always save the model after each epoch.")
    save_best_only: bool = Field(default=True, description="Only save the best model.")
    monitor: str = Field(
        default="valid_this_epoch_average_loss",
        description="The metric to monitor for saving best model.",
        examples=[
            "valid_this_epoch_average_loss",
            "valid_this_batch_average_loss",
            "train_this_epoch_average_loss",
            "train_this_batch_average_loss",
            "train_this_epoch_average_accuracy",
            "train_this_batch_average_accuracy",
            "valid_this_epoch_average_accuracy",
            "valid_this_batch_average_accuracy",
        ],
    )
    mode: str = Field(default="min", description="The mode to monitor for saving best model.", examples=["min", "max"])

    @field_validator("device", mode="before")
    @classmethod
    def set_device(cls: Type[TrainerConfig], v: str) -> torch.device:
        if v == "auto":
            return get_device()
        return torch.device(v)

    @field_validator("save_dir")
    @classmethod
    def set_and_create_timestamped_save_dir(cls: Type[TrainerConfig], v: str) -> str:
        timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        v = f"{v}/{timestamp}"
        Path(v).mkdir(parents=True, exist_ok=True)
        return v

    @model_validator(mode="after")
    def validate_autocast_config(self) -> TrainerConfig:
        use_amp = self.use_amp
        autocast_config = self.autocast_config
        if use_amp and not autocast_config["enabled"]:
            raise ValueError("If use_amp is True, autocast_config must be enabled.")

        for key, value in autocast_config.items():
            if key == "dtype" and value in PYTORCH_DTYPE_MAP:
                autocast_config[key] = PYTORCH_DTYPE_MAP[value]
        return self

    class Config:
        """Pydantic config."""

        arbitrary_types_allowed = True



import math
from typing import Any, Callable, Dict, Iterable, Optional, Tuple, Union

import torch
from typing_extensions import TypeAlias

# from torch.optim.optimizer import params_t as ParamsT

__all__ = ["AdamW"]

ParamsT: TypeAlias = Union[Iterable[torch.Tensor], Iterable[Dict[str, Any]]]


class AdamW(torch.optim.Optimizer):
    def __init__(
        self,
        params: ParamsT,
        lr: float = 1e-3,
        betas: Tuple[float, float] = (0.9, 0.999),
        eps: float = 1e-8,
        weight_decay: float = 0,
    ) -> None:
        self.validate_params(lr, betas, eps, weight_decay)

        defaults = {
            "lr": lr,
            "betas": betas,
            "eps": eps,
            "weight_decay": weight_decay,
        }
        super().__init__(params, defaults)

        # NOTE: state/param_groups are stated from pytorch optim source for clarity
        # self.state: DefaultDict[torch.Tensor, Any] = defaultdict(dict)
        # self.param_groups: List[Dict[str, Any]] = []

    @staticmethod
    def validate_params(lr: float, betas: Tuple[float, float], eps: float, weight_decay: float) -> None:
        if not 0.0 <= lr:
            raise ValueError(f"Invalid learning rate: {lr}, must be >= 0.")
        if not 0.0 <= eps:
            raise ValueError(f"Invalid epsilon value: {eps}, must be >= 0.")
        if not 0.0 <= weight_decay:
            raise ValueError(f"Invalid weight_decay value: {weight_decay}, must be >= 0.")
        if not 0.0 <= betas[0] < 1.0 and 0.0 <= betas[1] < 1.0:
            raise ValueError(f"Invalid beta values: {betas}, must be in [0, 1).")

    def step(self, closure: Optional[Callable[[], float]] = None) -> Optional[float]:  # type: ignore[override]
        loss = None
        if closure is not None:
            with torch.enable_grad():  # type: ignore[no-untyped-call]
                loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError("AdamW does not support sparse gradients")

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state["step"] = 0
                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(p, memory_format=torch.preserve_format)
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros_like(p, memory_format=torch.preserve_format)
                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                beta1, beta2 = group["betas"]

                state["step"] += 1

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)

                denom = exp_avg_sq.sqrt().add_(group["eps"])

                bias_correction1 = 1 - beta1 ** state["step"]
                bias_correction2 = 1 - beta2 ** state["step"]
                step_size = group["lr"] * math.sqrt(bias_correction2) / bias_correction1

                p.data.addcdiv_(exp_avg, denom, value=-step_size)

                if group["weight_decay"] != 0:
                    p.data.add_(p.data, alpha=-group["lr"] * group["weight_decay"])

        return loss



from enum import Enum


class TorchDtype(Enum):
    """Enum for PyTorch data types with memory size in bytes."""

    FLOAT32 = ("torch.float32", 4)
    FLOAT64 = ("torch.float64", 8)
    FLOAT16 = ("torch.float16", 2)
    BFLOAT16 = ("torch.bfloat16", 2)
    INT8 = ("torch.int8", 1)
    INT16 = ("torch.int16", 2)
    INT32 = ("torch.int32", 4)
    INT64 = ("torch.int64", 8)
    UINT8 = ("torch.uint8", 1)
    BOOL = ("torch.bool", 1)

    def __init__(self, dtype_str: str, size_in_bytes: int) -> None:
        self.dtype_str = dtype_str
        self.size_in_bytes = size_in_bytes

    def __str__(self) -> str:
        """Return the string representation of the enum value for dtype."""
        return self.dtype_str

    def get_size_in_bytes(self) -> int:
        """Return the size in bytes of the dtype."""
        return self.size_in_bytes



from pydantic import BaseModel, Field
from rich.pretty import pprint


class DistInfoPerProcess(BaseModel):
    """Assumes one process is one worker/gpu. Immutable and should only
    perform data validation."""

    master_addr: str = Field(
        ...,
        description="""
                    This is `MASTER_ADDR` which refers to the IP address (or hostname)
                    of the machine or node where the rank 0 process is running.
                    It acts as the reference point for all other nodes and GPUs
                    in the distributed setup. All other processes will connect
                    to this address for synchronization and communication.
                    """,
    )
    master_port: str = Field(
        ...,
        description="Denotes an available port on the `MASTER_ADDR` machine. "
        "All processes will use this port number for communication.",
    )

    nnodes: int = Field(
        ...,
        description="Number of nodes in the distributed setup.",
    )

    nproc_per_node: int = Field(
        ...,
        description="Number of processes/gpus per node in the distributed setup.",
    )

    node_rank: int = Field(
        ...,
        description="Rank of the current node in the distributed setup.",
    )

    world_size: int = Field(
        ...,
        description="Total number of processes/gpus in the distributed setup.",
    )

    # 3. Initialization of Process Group
    backend: str = Field(
        ...,
        description="Backend for distributed training.",
    )

    init_method: str = Field(
        "env://",
        description="Initialization method for distributed training.",
    )

    # 4. Others
    global_rank: int = Field(
        ...,
        description="Rank of the current process/gpu in the distributed setup.",
    )
    local_world_size: int = Field(
        ...,
        description="Total number of processes/gpus in the local node.",
    )
    local_rank: int = Field(
        ...,
        description="Rank of the current process/gpu in the local node.",
    )

    # 5. System info
    hostname: str = Field(
        ...,
        description="Hostname of the current node.",
    )
    process_id: int = Field(
        ...,
        description="Process ID of the current process.",
    )

    def pretty_print(self) -> None:
        pprint(self)



"""
The base module for the searching algorithms in the omnivault.dsa package. This
is the interface for the Strategy Design Pattern that will be implemented by the
concrete search strategies in strategies.py.

This module typically includes abstract base classes or common utility functions
that are utilized by the concrete strategy classes in strategies.py and the
context class in context.py.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Literal, Sequence, Union

from omnivault._types._alias import NonNegativeInt
from omnivault._types._generic import Real


class Search(ABC):
    """Base class for Search Strategies (Strategy Design Pattern)."""

    @abstractmethod
    def search(self, container: Sequence[Real], target: Real) -> Union[NonNegativeInt, Literal[-1]]:
        """Searches for the target in the container and returns the index of the
        target if found, otherwise returns -1."""



from __future__ import annotations

from abc import ABC
from pathlib import Path
from typing import Dict, List, Type, Union

import requests
from typing_extensions import override


class Vocabulary(ABC):
    """Base class for all vocabulary classes."""

    # Special tokens as class attributes
    BOS = "<BOS>"
    EOS = "<EOS>"
    PAD = "<PAD>"
    UNK = "<UNK>"

    def __init__(
        self,
        token_to_index: Dict[str, int],
        index_to_token: Dict[int, str],
    ) -> None:
        self.token_to_index = token_to_index
        self.index_to_token = index_to_token

    def __len__(self) -> int:
        """
        Returns the total number of unique tokens in the vocabulary.

        Returns
        -------
        int
            The number of tokens in the vocabulary.
        """
        return len(set(self.token_to_index.keys()))

    @property
    def vocab_size(self) -> int:
        """
        Returns the size of the vocabulary.

        Returns
        -------
        int
            The size of the vocabulary.
        """
        return len(self)

    @classmethod
    def from_tokens(cls: Type[Vocabulary], tokens: List[str]) -> Vocabulary:
        token_to_index = {token: idx for idx, token in enumerate(tokens)}
        index_to_token = {idx: token for token, idx in token_to_index.items()}
        return cls(token_to_index, index_to_token)


class AdderVocabulary(Vocabulary):
    # special tokens extended from Vocabulary
    ADD = "+"
    EQUAL = "="
    TOKENS = [
        "0",
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "8",
        "9",
        "+",
        "*",
        "-",
        "=",
        "<BOS>",
        "<EOS>",
        "<PAD>",
        "<UNK>",
    ]

    def __init__(self, token_to_index: Dict[str, int], index_to_token: Dict[int, str], num_digits: int) -> None:
        super().__init__(token_to_index, index_to_token)
        self.num_digits = num_digits

    @override
    @classmethod  # NOTE: why does overriding a classmethod with different signature not violate LSP?
    def from_tokens(cls: Type[AdderVocabulary], tokens: List[str], num_digits: int = 2) -> AdderVocabulary:
        token_to_index = {token: idx for idx, token in enumerate(tokens)}
        index_to_token = {idx: token for token, idx in token_to_index.items()}
        return cls(token_to_index, index_to_token, num_digits)


class TextCharacterVocabulary(Vocabulary):
    """
    A vocabulary class for character-level text processing. This class is designed
    to handle the encoding and decoding of characters in text data. It is particularly
    useful for tasks involving character-level models, such as GPT-style language models.

    The vocabulary consists of a set of unique characters found in a given text corpus.
    """

    PAD = "<PAD>"

    def __init__(self, token_to_index: Dict[str, int], index_to_token: Dict[int, str]) -> None:
        self.token_to_index = token_to_index
        self.index_to_token = index_to_token

    @staticmethod
    def _download(url: str, dataset_name: str, dest_folder: Path | str) -> Path:
        dest_folder_path = Path(dest_folder)

        dest_folder_path.mkdir(parents=True, exist_ok=True)

        filepath = dest_folder_path / f"{dataset_name}.txt"

        response = requests.get(url, stream=True)  # TODO: add timeout
        response.raise_for_status()

        with open(filepath, "wb") as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)

        return filepath

    @classmethod
    def from_corpus(cls: Type[TextCharacterVocabulary], corpus: str) -> TextCharacterVocabulary:
        vocabulary = sorted(set(corpus))
        token_to_index = {token: idx for idx, token in enumerate(vocabulary)}
        index_to_token = {idx: token for token, idx in token_to_index.items()}
        return cls(token_to_index, index_to_token)

    @classmethod
    def from_file(cls: Type[TextCharacterVocabulary], file_path: str | Path) -> TextCharacterVocabulary:
        with open(file_path, "r") as f:
            corpus = f.read()
        return cls.from_corpus(corpus)

    @classmethod
    def from_url(
        cls: Type[TextCharacterVocabulary], url: str, dataset_name: str, dest_folder: str | Path | None = None
    ) -> TextCharacterVocabulary:
        if not dest_folder:
            response = requests.get(url)
            response.raise_for_status()
            corpus = response.text
            return cls.from_corpus(corpus)

        file_path = cls._download(url, dataset_name, dest_folder)
        return cls.from_file(file_path)


Vocabularies = Union[AdderVocabulary, TextCharacterVocabulary]



from __future__ import annotations

from typing import Callable, Dict, Literal, Type, Union

import torch
from torch import nn

from omnivault.utils.config_management.dynamic import DynamicClassFactory

RegisteredCriterion = Literal["torch.nn.CrossEntropyLoss", "torch.nn.MSELoss"]
CRITERION_REGISTRY: Dict[RegisteredCriterion, Type[CriterionConfig]] = {}


def register_criterion(name: RegisteredCriterion) -> Callable[[Type[CriterionConfig]], Type[CriterionConfig]]:
    """
    Decorator factory for registering criterion configurations.

    Parameters
    ----------
    name : RegisteredCriterion
        The name to register the class under.

    Returns
    -------
    Callable[[Type[CriterionConfig]], Type[CriterionConfig]]
        A decorator that registers the criterion configuration class.
    """

    def register_criterion_cls(cls: Type[CriterionConfig]) -> Type[CriterionConfig]:
        if name in CRITERION_REGISTRY:
            raise ValueError(f"Cannot register duplicate criterion {name}")

        if not issubclass(cls, CriterionConfig):
            raise ValueError(f"Criterion (name={name}, class={cls.__name__}) must extend CriterionConfig")

        CRITERION_REGISTRY[name] = cls
        return cls

    return register_criterion_cls


class CriterionConfig(DynamicClassFactory[nn.Module]):
    """
    Configuration for creating PyTorch criterion instances dynamically.

    This class extends `DynamicClassFactory` to specifically handle the
    instantiation of PyTorch loss function classes based on provided configurations.

    Methods
    -------
    build() -> nn.Module
        Creates and returns a loss function instance with the specified parameters.
    """

    name: str

    class Config:
        """Pydantic config."""

        arbitrary_types_allowed = True


@register_criterion(name="torch.nn.CrossEntropyLoss")
class CrossEntropyLossConfig(CriterionConfig):
    weight: Union[torch.Tensor, None] = None
    size_average: Union[bool, None] = None
    ignore_index: int = -100
    reduction: str = "mean"
    label_smoothing: float = 0.0


@register_criterion(name="torch.nn.MSELoss")
class MSELossConfig(CriterionConfig):
    size_average: Union[bool, None] = None
    reduction: str = "mean"



from __future__ import annotations

from typing import Any, Dict, List, Literal, Tuple, TypeVar, Union, cast

import numpy as np
import numpy.typing as npt
import torch
from torch.utils.data import DataLoader, Dataset, Subset

from omnivault._types._alias import NotGiven
from omnivault._types._sentinel import NOT_GIVEN
from omnivault.transformer.core.tokenizer import AdderTokenizer, TextCharacterTokenizer
from omnivault.transformer.core.vocabulary import AdderVocabulary

# TODO: if both yield are same, then no point having the union of them except for semantics.
#       to fix this simply only use DatasetYield = Tuple[torch.LongTensor, torch.LongTensor, torch.BoolTensor, torch.BoolTensor]
#       and remove the union and the typevar.
AdderDatasetYield = Tuple[torch.LongTensor, torch.LongTensor, torch.BoolTensor, torch.BoolTensor]
TextCharacterDatasetYield = Tuple[torch.LongTensor, torch.LongTensor, torch.BoolTensor, torch.BoolTensor]
DatasetYield = Union[AdderDatasetYield, TextCharacterDatasetYield]
Dataset_co = TypeVar(
    "Dataset_co", bound=DatasetYield, covariant=True
)  # using covariant as pytorch Dataset is covariant in its type parameter


def get_batch(
    *,
    dataset: npt.NDArray[np.uint16],
    batch_size: int,
    context_length: int,
    generator: torch.Generator | None = None,
    device_type: Literal["cpu", "cuda"] = "cpu",  # exclude mps
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Generates a batch of input-output pairs from the given dataset.

    Parameters
    ----------
    dataset : numpy.ndarray
        The dataset to generate batches from. It should be a 1D NumPy array of type uint16.
    batch_size : int
        The number of samples in each batch.
    context_length : int
        The length of the context window for each sample.
    generator : torch.Generator, optional
        The PyTorch generator to use for generating random indices. If not provided,
        no seed is set.
    device_type : str, optional
        The device type to use for the generated tensors. Can be either "cpu" or "cuda".
        Defaults to "cpu".

    Returns
    -------
    Tuple[torch.Tensor, torch.Tensor]
        A tuple containing two tensors:
        - x: The input tensor of shape (batch_size, context_length).
        - y: The output tensor of shape (batch_size, context_length), where each element
             is shifted by 1 compared to the corresponding element in x.

    Notes
    -----
    - The function assumes that the dataset is a memory-mapped array to avoid memory leaks.
    - If the device type is "cuda", the function uses pinned memory for faster data transfer
      to the GPU.

    Examples
    --------
    >>> dataset = np.arange(100, dtype=np.uint16)
    >>> batch_size = 4
    >>> context_length = 10
    >>> x, y = get_batch(dataset=dataset, batch_size=batch_size, context_length=context_length)
    >>> x.shape
    torch.Size([4, 10])
    >>> y.shape
    torch.Size([4, 10])
    """
    # Source: Karpathy, We recreate np.memmap every batch to avoid a memory leak, as per
    # https://stackoverflow.com/questions/45132940/numpy-memmap-memory-usage-want-to-iterate-once/61472122#61472122

    # if not isinstance(dataset, np.memmap):
    #     raise ValueError("The dataset should be a memory-mapped array. Example: data = np.memmap('data.npy', dtype=np.uint16, mode='r')")

    device = torch.device("cuda") if device_type == "cuda" else torch.device("cpu")
    low, high = 0, len(dataset) - context_length
    size = (batch_size,)
    indices = torch.randint(low=low, high=high, size=size, generator=generator)

    x = torch.stack([torch.from_numpy((dataset[index : index + context_length]).astype(np.int64)) for index in indices])
    y = torch.stack(
        [torch.from_numpy((dataset[index + 1 : index + 1 + context_length]).astype(np.int64)) for index in indices]
    )
    if device_type == "cuda":
        # pin arrays x,y, which allows us to move them to GPU asynchronously (non_blocking=True)
        x, y = x.pin_memory().to(device, non_blocking=True), y.pin_memory().to(device, non_blocking=True)
    else:
        x, y = x.to(device), y.to(device)
    return x, y


class BaseDataset(Dataset[Dataset_co]):
    """Dataset base class, currently not filled in, but act as a conveyor for
    us to use Dataset[Dataset_co] as type hinting."""


# TODO: ideally splitting data should be done within the dataset class to
# speed up the process, as we only need to load the data that we need in memory.
# See Kapathy's https://github.com/karpathy/minGPT/tree/master/projects/adder.
class AdderDataset(BaseDataset[AdderDatasetYield]):
    """
    A Dataset class for encoding sequences for an addition problem.

    This dataset class takes a list of string representations of addition
    problems and a vocabulary object. It encodes the problems and prepares
    input and target tensors for model training.

    Parameters
    ----------
    data : List[str]
        A list of strings, each representing an addition problem, e.g., "15+57=072".
    tokenizer : AdderTokenizer
        A `AdderTokenizer` object used for encoding the strings into numerical tokens
        and decoding the numerical tokens back into strings.

    Attributes
    ----------
    vocabulary : Vocabulary
        A `Vocabulary` object used for encoding the strings into numerical tokens.
    equal_token_id : int
        The numerical token ID for the equal sign.
    pad_token_id : int
        The numerical token ID for the padding token.
    """

    def __init__(self, data: List[str], tokenizer: AdderTokenizer) -> None:
        super().__init__()

        self.data = data
        self.tokenizer = tokenizer
        self.vocabulary = tokenizer.vocabulary

        self.equal_token_id: int = self.vocabulary.token_to_index[AdderVocabulary.EQUAL]
        self.pad_token_id: int = self.vocabulary.token_to_index[AdderVocabulary.PAD]

    def __len__(self) -> int:
        return len(self.data)

    def construct_future_mask(self, seq_len: int) -> torch.BoolTensor:
        future_mask = torch.tril(torch.ones((seq_len, seq_len), dtype=torch.bool), diagonal=0).to(torch.bool)
        future_mask = future_mask.contiguous()
        return torch.BoolTensor(future_mask)

    def construct_padding_mask(self, input_sequence: torch.Tensor) -> torch.BoolTensor:
        padding_mask = input_sequence != self.pad_token_id
        return torch.BoolTensor(padding_mask)

    def construct_input_tensor(self, input_sequence: torch.Tensor) -> torch.LongTensor:
        # Returns all but the last token
        return torch.LongTensor(input_sequence[:-1])

    def construct_target_tensor(self, input_sequence: torch.Tensor) -> torch.LongTensor:
        # Masks out tokens before the equal sign
        # TODO: is clone replaceable? it is not removable for now due to mutation.
        target = input_sequence.clone()
        where_equal_index = torch.where(input_sequence == self.equal_token_id)[0].item()
        where_equal_index = int(where_equal_index)  # to appease mypy lol
        target[: where_equal_index + 1] = self.pad_token_id
        return torch.LongTensor(target[1:])

    def __getitem__(self, index: int) -> AdderDatasetYield:
        """
        data = ["15+57=072", "92+00=092", "95+53=148", "15+10=025"]
        getitem selects one index randomly, say 2, to obtain
        data[2] -> "95+53=148"
        we need split to input and target by slicing
        x takes all but last token
        y takes all but first token
        """
        raw_sequence: str = self.data[index]
        encoded_sequence: List[int] = self.tokenizer.encode(raw_sequence)

        input_sequence: torch.LongTensor = torch.tensor(encoded_sequence, dtype=torch.long)  # type: ignore[assignment]

        input = self.construct_input_tensor(input_sequence)  # x
        target = self.construct_target_tensor(input_sequence)  # y
        padding_mask = self.construct_padding_mask(input)
        future_mask = self.construct_future_mask(input.size(0))
        return cast(AdderDatasetYield, (input, target, padding_mask, future_mask))  # TODO: really mypy?


class TextCharacterDataset(BaseDataset[TextCharacterDatasetYield]):
    """
    A Dataset class for encoding sequences from a text corpus at the character
    level.

    This dataset class takes a text corpus and a vocabulary object. It encodes the
    text corpus into numerical tokens and prepares input and target tensors for
    model training, based on the specified context length.

    Parameters
    ----------
    text_corpus : str
        A string representing the entire text corpus.
    vocabulary : List[str]
        A list of unique characters representing the vocabulary.
    context_length : int
        The length of the context window used for creating each training example.
    """

    def __init__(self, corpus: str, context_length: int, tokenizer: TextCharacterTokenizer) -> None:
        super().__init__()

        self.corpus = corpus
        self.context_length = context_length
        self.tokenizer = tokenizer
        self.vocabulary = tokenizer.vocabulary

    def construct_future_mask(self, seq_len: int) -> torch.BoolTensor:
        future_mask = torch.tril(torch.ones((seq_len, seq_len), dtype=torch.bool), diagonal=0).to(torch.bool)
        future_mask = future_mask.contiguous()
        return torch.BoolTensor(future_mask)

    @property
    def corpus_size(self) -> int:
        return len(self.corpus)

    def __len__(self) -> int:
        return len(self.corpus) - self.context_length

    def __getitem__(self, index: int) -> TextCharacterDatasetYield:
        """
        Retrieves a training example based on the specified index.

        The method selects a context window from the text corpus and encodes it into numerical tokens.
        It then splits the encoded context into input and target tensors for training.

        Parameters
        ----------
        index : int
            The index at which to start the context window.

        Returns
        -------
        Tuple[torch.LongTensor, torch.LongTensor]
            A tuple containing the input and target tensors.
        """
        context = self.corpus[index : index + self.context_length + 1]
        context_encoded = self.tokenizer.encode(context)
        x = torch.tensor(context_encoded[:-1], dtype=torch.long)
        y = torch.tensor(context_encoded[1:], dtype=torch.long)

        padding_mask = torch.ones_like(x, dtype=torch.bool)

        seq_len = x.size(0)
        assert seq_len == self.context_length, f"seq_len {seq_len} != context_length {self.context_length}"
        future_mask = self.construct_future_mask(seq_len)

        return cast(TextCharacterDatasetYield, (x, y, padding_mask, future_mask))


def construct_dummy_batch_future_masks(batch_size: int, seq_len: int) -> torch.BoolTensor:
    """Broadcast future mask from shape (L, L) to (B, L, L) then (B, 1, L, L)."""
    # Create a lower triangular mask for a single sequence
    future_mask = torch.tril(torch.ones((seq_len, seq_len), dtype=torch.bool), diagonal=0).to(torch.bool)
    future_mask = future_mask.contiguous()
    # broadcast future mask from shape (L, L) to (B, L, L)
    future_masks = future_mask.unsqueeze(0).expand(batch_size, -1, -1)
    # broadcast future mask from shape (B, L, L) to (B, 1, L, L)
    future_masks = future_masks.unsqueeze(1)
    return torch.BoolTensor(future_masks)


def construct_dummy_batch_target_padding_masks(batch_size: int, seq_len: int) -> torch.BoolTensor:
    """Construct a dummy batch of target padding masks of shape (B, 1, L, L) which
    assumes there is no padding token involved."""

    return torch.BoolTensor(torch.ones((batch_size, 1, seq_len, seq_len), dtype=torch.bool))


def collate_fn(
    batch: List[DatasetYield],
    batch_first: bool = True,
    pad_token_id: int = 0,
) -> DatasetYield:
    """
    Collates a batch of data into padded tensors for input, target, and masks.

    This function takes a batch of data produced by `AdderDataset` instances and
    pads the sequences to the same length, preparing them for batched processing
    in models. It also reshapes padding and future masks to match the batch
    dimensions.

    Parameters
    ----------
    batch : List[DatasetYield]
        A batch of data, where each item is a tuple containing input, target,
        padding mask, and future mask tensors.
    batch_first : bool, default=True
        If `batch_first=True`, the resulting tensor `inputs_padded` will have a shape of
        `(batch_size, max_seq_len)`, where `batch_size` is the number of samples in the
        batch and `max_seq_len` is the length of the longest sequence in the batch. If
        `batch_first=False`, the shape will be `(max_seq_len, batch_size)`. Note here
        `max_seq_len` is similar to `context_length` in our terminology.
    pad_token_id : int, default=0
        The `padding_value` parameter specifies the value to use for padding shorter
        sequences. `pad_token_id` is typically used here, which should correspond to the
        padding token's ID in your vocabulary.

    Returns
    -------
    DatasetYield
        A tuple containing the following tensors:
        - Padded input tensor
        - Padded target tensor
        - Reshaped padding mask tensor
        - Reshaped and expanded future mask tensor

    Notes
    -----
    The function assumes that all sequences in the batch are of the same type
    and can be padded with the same `pad_token_id`.
    """
    # omega confused during zipping so put here for clarity
    # that when you unzip a batch it becomes a tuple.
    # inputs: Tuple[torch.Tensor, ...]
    # targets: Tuple[torch.Tensor, ...]
    # padding_masks: Tuple[torch.Tensor, ...]
    # future_masks: Tuple[torch.Tensor, ...]

    inputs, targets, padding_masks, future_masks = zip(*batch)

    # Padding sequences to the same length: convert to list to appease typing
    # of pad_sequence as it expects a list of tensors or tensors, not tuple.
    inputs_padded: torch.Tensor = torch.nn.utils.rnn.pad_sequence(
        list(inputs), batch_first=batch_first, padding_value=pad_token_id
    )
    targets_padded: torch.Tensor = torch.nn.utils.rnn.pad_sequence(
        list(targets), batch_first=batch_first, padding_value=pad_token_id
    )
    padding_masks_padded: torch.Tensor = torch.nn.utils.rnn.pad_sequence(
        list(padding_masks), batch_first=batch_first, padding_value=pad_token_id
    )
    # Reshaping padding masks
    batch_size, seq_len = inputs_padded.size(0), inputs_padded.size(1)

    # padding_masks before view has shape: (batch_size, seq_len)
    # we want it to be (B, L, L) then (B, 1, L, L)
    padding_masks_padded_and_expanded = padding_masks_padded.view(batch_size, 1, 1, seq_len).expand(
        batch_size, 1, seq_len, seq_len
    )

    # future mask has shape (L, L) but we want it to be (B, L, L) then (B, 1, L, L)
    future_masks: torch.BoolTensor = torch.stack(future_masks)  # type: ignore[assignment,no-redef]

    future_masks_expanded = future_masks.expand(batch_size, -1, -1).unsqueeze(1)  # type: ignore[attr-defined]
    return cast(
        AdderDatasetYield, (inputs_padded, targets_padded, padding_masks_padded_and_expanded, future_masks_expanded)
    )


def create_loader(
    dataset: Dataset[Dataset_co],
    loader_config: Dict[str, Any],
    collate_fn_config: Dict[str, Any] | NotGiven = NOT_GIVEN,
) -> DataLoader[Dataset_co]:
    if collate_fn_config is NOT_GIVEN:
        return DataLoader(
            dataset=dataset,
            **loader_config,
        )
    return DataLoader(
        dataset=dataset,
        collate_fn=lambda batch: collate_fn(batch, **collate_fn_config),  # type: ignore[arg-type]
        **loader_config,
    )


def split_dataset(
    dataset: Dataset[Dataset_co], split: List[float], seed: int
) -> Tuple[
    Subset[DatasetYield], Subset[DatasetYield], Subset[DatasetYield]
]:  # TODO: unclean since it should return AdderDataset[AdderDatasetYield] but mypy is not happy
    train_dataset, val_dataset, test_dataset = torch.utils.data.random_split(
        dataset,
        lengths=split,
        generator=torch.Generator().manual_seed(seed),
    )
    return train_dataset, val_dataset, test_dataset



from typing import Any

from pydantic import BaseModel

__all__ = ["MaybeConstant"]


class MaybeConstant(BaseModel):
    """The maybe constant config that allows arbitrary fields. Not type safe
    for sure! So have to use type ignore if mypy cannot locate dynamically
    generated fields."""

    def __init__(self, **arbitrary: Any) -> None:
        super().__init__(**arbitrary)

    class Config:
        """Pydantic config."""

        arbitrary_types_allowed = True
        extra = "allow"



from __future__ import annotations

from typing import Generator, List

from omnivault._types._generic import T
from omnivault.dsa.stack.base import Stack


class StackList(Stack[T]):
    """Creates a stack that uses python's default list as the underlying
    data structure.

    Note
    ----
    Methods are ordered with
    dunder/magic/property -> public -> private -> static/class.

    Attributes
    ----------
    _stack_items : List[T]
        The list that stores the items in the stack. We treat the end of the
        list as the top of the stack.
    """

    def __len__(self) -> int:
        """Return the size of the stack."""
        return len(self.stack_items)

    def __iter__(self) -> Generator[T, None, None]:
        """Iterate over the stack items.

        Note
        ----
        If we return self, then we need to define `__next__`
        to make it an iterator. Else, python handles the
        `__next__` method for us if `__iter__` returns an
        iterator.

        ```python
        def __next__(self) -> StackList[T]:
            if self.is_empty():
                raise StopIteration
            return self.pop()
        ```

        Returns
        -------
        StackList[T]
            The stack.
        """

        while not self.is_empty():
            yield self.pop()

    @property
    def stack_items(self) -> List[T]:
        """Read only property for the stack items."""
        return self._stack_items

    @property
    def size(self) -> int:
        """Return the size of the stack.

        Note
        ----
        When you call `len(self)` from within the class, it will call internally
        `self.__len__()` (`StackList.__len__()`) which will return the size of
        the stack.

        Returns
        -------
        int
            The size of the stack.
        """
        return len(self)

    def is_empty(self) -> bool:
        """Check if stack is empty.

        Returns
        -------
        bool
            True if stack is empty, False otherwise.
        """
        return not self.stack_items

    def peek(self) -> T:
        """Return the top most item in the stack without modifying the stack.

        This is different from pop in that it does not remove the item from the
        stack.

        Returns
        -------
        T
            The top most item in the stack.
        """
        return self.stack_items[-1]

    def pop(self) -> T:
        """Pop an item from the top of the stack.

        In this implementation, the item at the end of the list is returned
        and removed. We are using the list's pop method to do this.

        Raises
        ------
        (Exception): If stack is empty.

        Returns
        -------
        T
            The top most item in the stack.
        """
        if self.is_empty():
            raise Exception("Stack is empty")
        return self.stack_items.pop()

    def push(self, item: T) -> None:
        """Push an item on top of the stack.

        In this implementation, the item is appended to the end of the list.

        Parameters
        ----------
        item : T
            The current item pushed into the stack.
        """
        self.stack_items.append(item)



"""
This module provides the FigureManager class, a utility for managing
matplotlib figures and axes. It's designed to abstract and simplify the
process of figure and axes creation and customization in matplotlib, making
it easier to create and manage plots in a reusable manner.
"""

from typing import Any, Dict, Optional, Union

import matplotlib.pyplot as plt


class FigureManager:
    """
    A utility class for managing matplotlib figures and axes.

    The class provides an interface to create a figure and an axes object,
    either by using existing ones or creating new ones. It also allows for
    the application of custom settings to the axes object through a
    dictionary of attributes and their corresponding parameters.
    """

    def __init__(
        self,
        fig: Optional[plt.Figure] = None,
        ax: Optional[plt.Axes] = None,
        ax_kwargs: Optional[Dict[str, Dict[str, Any]]] = None,
    ) -> None:
        # fmt: off
        self.fig       = fig or plt.gcf()
        self.ax        = ax or plt.gca()
        self.ax_kwargs = ax_kwargs or {}
        # fmt: on

        self._apply_ax_customizations()

    def _apply_ax_customizations(self) -> None:
        """
        Apply custom settings to the axes object.

        This method iterates through the `ax_kwargs` dictionary and applies
        each attribute and its parameters to the axes object.
        """
        for ax_attr, ax_params in self.ax_kwargs.items():
            getattr(self.ax, ax_attr)(**ax_params)

    def show(self) -> None:
        """Show the figure."""
        if self.fig:
            self.fig.show()
        else:
            plt.show()  # type: ignore[no-untyped-call]

    def save(
        self,
        path: str,
        *,
        dpi: Union[float, str] = "figure",
        format: str = "svg",
        **kwargs: Dict[str, Any],
    ) -> None:
        self.fig.savefig(path, dpi=dpi, format=format, **kwargs)  # type: ignore[arg-type]



from __future__ import annotations

from typing import Dict, List

import numpy as np
import torch
from scipy.special import softmax
from sklearn.metrics import (
    accuracy_score,
    auc,
    average_precision_score,
    brier_score_loss,
    confusion_matrix,
    f1_score,
    log_loss,
    precision_recall_curve,
    precision_score,
    recall_score,
    roc_auc_score,
    roc_curve,
)
from transformers import PreTrainedModel, PreTrainedTokenizerBase, PreTrainedTokenizerFast
from transformers.trainer_utils import EvalPrediction


def maybe_resize_token_embeddings(
    model: PreTrainedModel, tokenizer: PreTrainedTokenizerBase | PreTrainedTokenizerFast
) -> bool:
    """Returns True if the token embeddings need to be resized to match the tokenizer's vocabulary size.

    Parameters
    ----------
    model : PreTrainedModel
        The model to check for token embeddings.
    tokenizer : PreTrainedTokenizerBase | PreTrainedTokenizerFast
        The tokenizer to check for vocabulary size.

    Returns
    -------
    bool
        Returns True if the token embeddings need to be resized, False otherwise.
    """
    try:
        embedding_module: torch.nn.Module = model.get_input_embeddings()
        embedding_size: int = embedding_module.weight.shape[0]
    except AttributeError as exc:
        raise AttributeError("`weight` attribute not found in the model's input embeddings. ") from exc
    tokenizer_vocab_size = len(tokenizer)
    return embedding_size != tokenizer_vocab_size


def smart_tokenizer_and_embedding_resize(
    model: PreTrainedModel,
    tokenizer: PreTrainedTokenizerBase | PreTrainedTokenizerFast,
    special_tokens_dict: Dict[str, int],
) -> None:
    """Resize tokenizer and embedding.

    Note: This is the unoptimized version that may make your embedding size
    not be divisible by 64. This is also referenced from `qlora/qlora.py`
    https://github.com/artidoro/qlora/blob/7f4e95a68dc076bea9b3a413d2b512eca6d004e5/qlora.py#L425.

    Parameters
    ----------
    model : PreTrainedModel
        The model to check for token embeddings.
    tokenizer : PreTrainedTokenizerBase | PreTrainedTokenizerFast
        The tokenizer to check for vocabulary size.
    special_tokens_dict : Dict[str, int]
        Dictionary containing special tokens to add.
    """
    num_new_tokens = tokenizer.add_special_tokens(special_tokens_dict)
    model.resize_token_embeddings(len(tokenizer))

    if num_new_tokens > 0:
        input_embeddings_data = model.get_input_embeddings().weight.data
        output_embeddings_data = model.get_output_embeddings().weight.data

        input_embeddings_avg = input_embeddings_data[:-num_new_tokens].mean(dim=0, keepdim=True)
        output_embeddings_avg = output_embeddings_data[:-num_new_tokens].mean(dim=0, keepdim=True)

        input_embeddings_data[-num_new_tokens:] = input_embeddings_avg
        output_embeddings_data[-num_new_tokens:] = output_embeddings_avg


def compute_metrics_for_single_label_classification(eval_prediction: EvalPrediction) -> Dict[str, float | List[float]]:
    logits, labels = eval_prediction.predictions, eval_prediction.label_ids
    probs = softmax(logits, axis=-1)

    num_classes = logits.shape[1]
    preds = np.argmax(probs, axis=1)

    metrics = {
        "eval_log_loss": log_loss(labels, probs),
        "eval_accuracy": accuracy_score(labels, preds),
        "eval_precision_macro": precision_score(labels, preds, average="macro", zero_division=0),
        "eval_recall_macro": recall_score(labels, preds, average="macro", zero_division=0),
        "eval_f1_score_macro": f1_score(labels, preds, average="macro", zero_division=0),
        "eval_precision_micro": precision_score(labels, preds, average="micro", zero_division=0),
        "eval_recall_micro": recall_score(labels, preds, average="micro", zero_division=0),
        "eval_f1_score_micro": f1_score(labels, preds, average="micro", zero_division=0),
        "eval_confusion_matrix": confusion_matrix(labels, preds).tolist(),
        "eval_roc_auc": roc_auc_score(labels, probs, multi_class="ovr"),
        "eval_pr_auc": average_precision_score(labels, probs, average="macro"),
    }

    if num_classes == 2:
        metrics["eval_brier_score"] = brier_score_loss(labels, probs[:, 1], pos_label=1)
    else:
        brier_scores = [brier_score_loss(labels == i, probs[:, i]) for i in range(num_classes)]
        metrics["eval_brier_score"] = np.mean(brier_scores)

    if num_classes > 2:
        for class_index in range(num_classes):
            fpr, tpr, _ = roc_curve(labels == class_index, probs[:, class_index])
            roc_auc = auc(fpr, tpr)
            precision, recall, _ = precision_recall_curve(labels == class_index, probs[:, class_index])
            pr_auc = auc(recall, precision)
            metrics[f"eval_roc_auc_class_{class_index}"] = roc_auc
            metrics[f"eval_pr_auc_class_{class_index}"] = pr_auc

    return metrics



from __future__ import annotations

import copy
import logging
import sys
import time
import warnings

import pandas as pd
import torch
from hydra.utils import instantiate
from omegaconf import DictConfig, ListConfig
from omegaconf import OmegaConf as om
from tqdm import tqdm

from omnivault._types._alias import Missing
from omnivault._types._sentinel import MISSING
from omnivault.core.logger import RichLogger
from omnivault.transformer.config.composer import Composer, DataConfig
from omnivault.transformer.config.constants import MaybeConstant
from omnivault.transformer.config.criterion import CRITERION_REGISTRY
from omnivault.transformer.config.decoder import DecoderConfig
from omnivault.transformer.config.generator import GeneratorConfig
from omnivault.transformer.config.global_ import MaybeGlobal
from omnivault.transformer.config.logger import LoggerConfig
from omnivault.transformer.config.optim import OPTIMIZER_REGISTRY
from omnivault.transformer.config.scheduler import SCHEDULER_REGISTRY, LambdaLRConfig
from omnivault.transformer.config.trainer import TrainerConfig
from omnivault.transformer.core.dataset import AdderDataset, create_loader, split_dataset
from omnivault.transformer.core.optim import apply_weight_decay_to_different_param_groups
from omnivault.transformer.core.scheduler import noam_lr_decay
from omnivault.transformer.core.state import State
from omnivault.transformer.core.tokenizer import AdderTokenizer
from omnivault.transformer.core.trainer import Trainer, TrainerEvent
from omnivault.transformer.core.vocabulary import AdderVocabulary
from omnivault.transformer.decoder.core import GPTDecoder
from omnivault.transformer.utils.general_utils import create_directory, download_file, validate_and_cleanup
from omnivault.transformer.utils.visualization import save_plot_history
from omnivault.utils.config_management.omegaconf import load_yaml_config, merge_configs
from omnivault.utils.reproducibility.seed import seed_all

warnings.filterwarnings("ignore", category=UserWarning)  # usually related to deterministic behavior of pytorch

# TODO: I have a callable instead of _target_ field for me to use importlib to parse.
# so maybe consider using my own code base


@torch.no_grad()
def evaluate_and_generate_on_valid_epoch_end(
    trainer: Trainer,
    num_batches_to_eval: int | None = None,
) -> None:
    generator_config = trainer.composer.generator
    assert (
        generator_config.max_tokens == trainer.composer.constants.NUM_DIGITS + 1 + 1  # type: ignore[attr-defined]
    ), "In this dataset, the max tokens to generate is fixed and derived from the number of digits. If we add two 2-digits together, it does not make sense for us to keep generating since the max digits for answer is 3 digits, with an optional `<EOS>` token if it is in our vocabulary."
    assert generator_config.greedy is True, "We should use greedy generation for this task in particular."

    vocabulary = trainer.state.vocabulary
    assert isinstance(vocabulary, AdderVocabulary)

    tokenizer = trainer.state.tokenizer
    assert isinstance(tokenizer, AdderTokenizer)

    EQUAL = vocabulary.token_to_index[vocabulary.EQUAL]
    EOS = vocabulary.token_to_index[vocabulary.EOS]

    model = trainer.model_or_module
    model.eval()

    dataloader = trainer.test_loader
    assert dataloader is not None
    total_samples = 0
    num_batches = len(dataloader)

    total_correct_across_samples = 0
    progress_bar = tqdm(
        enumerate(dataloader, start=1), total=num_batches, desc="Evaluating and Generation.", leave=False
    )

    all_predictions = []
    for _batch_index, batch in progress_bar:
        inputs, _, _, _ = batch
        inputs = inputs.to(trainer.device)

        # example of 1 batch with 1 sample:
        # inputs:   [14, 9, 8, 10, 9, 9, 13, 1, 9, 7]
        # targets:  [    9, 8, 10, 9, 9, 13, 1, 9, 7, 15]
        # equation: [14, 9, 8, 10, 9, 9, 13, 1, 9, 7, 15]
        # equal_index: 6'th position
        # starting_tokens: [14, 9, 8, 10, 9, 9, 13] -> <BOS>98+99=
        # generated_tokens: [14, 9, 8, 10, 9, 9, 13, 1, 9, 7, 15]
        # generated_tokens_decoded: 98+99=197

        batch_correct_predictions = 0
        batch_size = inputs.size(0)

        # [batch_size, 1]
        # [[15], [15], ...]
        eos_tokens_batch = torch.full((batch_size, 1), EOS, dtype=torch.long, device=trainer.device)

        # [batch_size, context_length] = [batch_size, 11] including <BOS> and <EOS>
        # [[14,  1,  3, 10,  4,  8, 13,  0,  6,  1, 15], [14,  1,  6, 10,  5,  5, 13,  0,  7,  1, 15]]
        equations = torch.cat((inputs, eos_tokens_batch), dim=1)  # this is the full equation with answer
        equations_decoded = tokenizer.decode_batch(encoded_sequences=equations, remove_special_tokens=True)

        # [batch_size,] = [6, 6, ...]
        equal_indices = torch.where(equations == EQUAL)[1]  # .view(batch_size, -1)[:, 0]

        # [batch_size, 7] because each starting token is <BOS>AB+CD=
        starting_tokens_batch = torch.zeros(
            batch_size, int(equal_indices[0].item()) + 1, dtype=torch.long, device=trainer.device
        )
        for i in range(batch_size):
            starting_tokens_batch[i] = equations[i, : equal_indices[i] + 1]

        # [batch_size, context_length] = [batch_size, 11] including <BOS> and <EOS>
        generated_tokens_batch = model.generate(
            starting_tokens=starting_tokens_batch,
            **generator_config.model_dump(mode="python"),
        )
        generated_tokens_decoded = tokenizer.decode_batch(
            encoded_sequences=generated_tokens_batch, remove_special_tokens=True
        )
        is_correct_batch = torch.all(torch.eq(generated_tokens_batch, equations), dim=1)

        batch_correct_predictions = torch.sum(is_correct_batch).item()  # type: ignore[assignment]
        total_correct_across_samples += batch_correct_predictions
        batch_accuracy = batch_correct_predictions / batch_size
        progress_bar.set_postfix_str(f"accuracy: {batch_accuracy:.4f}")

        for _sample_index in range(batch_size):
            total_samples += 1
            is_correct = is_correct_batch[_sample_index]
            all_predictions.append(
                {
                    "epoch": trainer.epoch_index,
                    "batch_index": _batch_index,
                    "sample_index": _sample_index,
                    "equation": equations_decoded[_sample_index],
                    "generated": generated_tokens_decoded[_sample_index],
                    "correct": is_correct.item(),
                }
            )

        if num_batches_to_eval and _batch_index >= num_batches_to_eval:
            trainer.logger.info("Early stopping evaluation.")
            break

    accuracy = total_correct_across_samples / total_samples
    df = pd.DataFrame(all_predictions)

    trainer.logger.info("%-32s %d/%d", "Correct/Total Samples:", total_correct_across_samples, total_samples)
    trainer.logger.info("%-32s %s", "Eval Accuracy:", accuracy)

    df_str = df.to_string(index=False)
    trainer.logger.info("\n%s", df_str)


def main(cfg: DictConfig | ListConfig) -> None:
    """Main driver."""
    seed_all(cfg.global_.seed)

    constants = MaybeConstant(**cfg.constants) if cfg.constants is not None else MaybeConstant()
    logger_pydantic_config = LoggerConfig(**cfg.logger)
    global_ = MaybeGlobal(**cfg.global_)
    data = DataConfig(**cfg.data)
    trainer_config = TrainerConfig(**cfg.trainer)
    generator_config = GeneratorConfig(**cfg.generator)

    # logger
    logger = RichLogger(**logger_pydantic_config.model_dump(mode="python")).logger
    assert isinstance(logger, logging.Logger)

    create_directory(data.dataset_dir)
    download_file(url=data.dataset_url, output_path=data.dataset_path)

    vocabulary = AdderVocabulary.from_tokens(tokens=constants.TOKENS, num_digits=constants.NUM_DIGITS)  # type: ignore[attr-defined]
    tokenizer = AdderTokenizer(vocabulary=vocabulary)

    # assign back model.vocab_size from ??? to vocabulary.vocab_size
    cfg.model.vocab_size = vocabulary.vocab_size
    model_config = instantiate(cfg.model)
    model_pydantic_config = DecoderConfig(**model_config)

    optimizer_config_cls = OPTIMIZER_REGISTRY[cfg.optimizer.name]
    optimizer_pydantic_config = optimizer_config_cls(**cfg.optimizer)

    criterion_config_cls = CRITERION_REGISTRY[cfg.criterion.name]
    criterion_pydantic_config = criterion_config_cls(**cfg.criterion)

    composer = Composer(
        constants=constants,
        logger=logger_pydantic_config,
        global_=global_,
        data=data,
        model=model_pydantic_config,
        optimizer=optimizer_pydantic_config,
        criterion=criterion_pydantic_config,
        trainer=trainer_config,
        generator=generator_config,
    )
    assert composer.model is not MISSING and not isinstance(composer.model, Missing)
    assert composer.optimizer is not MISSING and not isinstance(composer.optimizer, Missing)
    assert composer.criterion is not MISSING and not isinstance(composer.criterion, Missing)

    # TODO: consider classmethod from file_path
    assert composer.data.dataset_path is not None
    with open(composer.data.dataset_path, "r") as file:
        sequences = [line.strip() for line in file]

    dataset = AdderDataset(data=sequences, tokenizer=tokenizer)
    if composer.data.split:
        train_dataset, valid_dataset, test_dataset = split_dataset(
            dataset=dataset, split=composer.data.split, seed=composer.global_.seed
        )
    else:
        # no need to cater to mypy as either Subset or Dataset is fine.
        train_dataset = dataset  # type: ignore[assignment]

    # you do these asserts to make sure that the loaders are not None
    # because create loader expects non-None loaders and collate_fn.
    # if you don't do these asserts, mypy cannot guarantee that the loaders are not None
    # so they cannot infer properly.
    assert composer.data.train_loader is not None
    assert composer.data.valid_loader is not None
    assert composer.data.test_loader is not None
    assert composer.data.collate_fn is not None

    train_loader = create_loader(
        dataset=train_dataset,
        loader_config=composer.data.train_loader,
        collate_fn_config=composer.data.collate_fn,
    )

    if valid_dataset is not None:
        valid_loader = create_loader(
            dataset=valid_dataset,
            loader_config=composer.data.valid_loader,
            collate_fn_config=composer.data.collate_fn,
        )

    if test_dataset is not None:
        test_loader = create_loader(  # noqa: F841
            dataset=test_dataset,
            loader_config=composer.data.test_loader,
            collate_fn_config=composer.data.collate_fn,
        )

    # Create model
    model = GPTDecoder(model_pydantic_config)
    model = model.to(device=composer.trainer.device, dtype=next(model.parameters()).dtype, non_blocking=True)

    # Create optimizer based on model parameters
    if composer.trainer.apply_weight_decay_to_different_param_groups:
        assert hasattr(composer.optimizer, "weight_decay")
        optimizer = optimizer_pydantic_config.build(
            params=apply_weight_decay_to_different_param_groups(
                model=model, weight_decay=composer.optimizer.weight_decay
            )
        )
    else:
        optimizer = optimizer_pydantic_config.build(params=model.parameters())

    # Create criterion
    criterion = criterion_pydantic_config.create_instance()
    assert criterion.ignore_index == vocabulary.token_to_index[vocabulary.PAD]

    # Create Scheduler noam
    # TODO: this part is hardcoded in a way since we are using LambdaLR.
    # I do not have time to make it more "automated" so this is anti-config-pattern.
    warmup_steps = 3 * len(train_loader)

    # lr first increases in the warmup steps, and then decays
    noam = lambda step: noam_lr_decay(step, d_model=composer.model.d_model, warmup_steps=warmup_steps)  # noqa: E731

    scheduler_config_cls = SCHEDULER_REGISTRY[cfg.scheduler.name]

    if issubclass(scheduler_config_cls, LambdaLRConfig):
        scheduler_pydantic_config = scheduler_config_cls(lr_lambda=noam, **cfg.scheduler)
    else:
        scheduler_pydantic_config = scheduler_config_cls(**cfg.scheduler)  # type: ignore[assignment]

    assert composer.scheduler is MISSING  # now it is MISSING for us to fill up.
    composer.scheduler = scheduler_pydantic_config
    scheduler = scheduler_pydantic_config.build(optimizer=optimizer)

    composer.pretty_print()
    time.sleep(1)

    state = State(
        model=model,
        criterion=criterion,
        optimizer=optimizer,
        scheduler=scheduler,
        vocabulary=vocabulary,
        tokenizer=tokenizer,
    )
    state.pretty_print()
    time.sleep(1)

    device = composer.trainer.device

    # train
    trainer = Trainer(
        state=state,
        composer=composer,
        logger=logger,
        device=device,  # type: ignore[arg-type]
    )
    trainer.add_callback(
        TrainerEvent.ON_VALID_EPOCH_END,
        lambda trainer: evaluate_and_generate_on_valid_epoch_end(trainer, num_batches_to_eval=None),
    )
    _trained_state = trainer.fit(train_loader=train_loader, valid_loader=valid_loader, test_loader=test_loader)
    _trained_state.pretty_print()
    history = _trained_state.history
    _ = save_plot_history(history, plot=False, save_path=f"{composer.trainer.save_dir}/history.png")

    loaded_state = State.load_snapshots(
        filepath=trainer.best_checkpoint_path,
        device=device,  # type: ignore[arg-type]
        model=copy.deepcopy(model),
        criterion=criterion,
        optimizer=optimizer,
        scheduler=scheduler,
    )

    validate_and_cleanup(
        state_1=_trained_state,
        state_2=loaded_state,
        objects=["model", "criterion", "optimizer", "scheduler", "tokenizer", "vocabulary", "trainer"],
        logger=None,
    )


if __name__ == "__main__":
    yaml_path = sys.argv[1]
    args_list = sys.argv[2:]

    yaml_cfg = load_yaml_config(yaml_path)
    cfg = merge_configs(yaml_cfg, args_list)
    om.resolve(cfg)  # inplace ops

    main(cfg)
